/* Modular diagnostic dump functions.

   Copyright 2001, 2002, 2007 Kevin Ryde

   This file is part of MPM.

   MPM is free software; you can redistribute it and/or modify it under the
   terms of the GNU General Public License as published by the Free Software
   Foundation; either version 2, or (at your option) any later version.

   MPM is distributed in the hope that it will be useful, but WITHOUT ANY
   WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
   FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
   details.

   You should have received a copy of the GNU General Public License along
   with MPM; see the file COPYING.  If not, write to the Free Software
   Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301
   USA.  */

#include <stdio.h>
#include "gmp.h"
#include "gmp-impl.h"
#include "longlong.h"
#include "tests/tests.h"

#include "mpm.h"
#include "mpm-impl.h"


void
mpr_dump (mpr_srcptr r, mpm_srcptr m)
{
  gmp_printf ("r %p: %#NX\n", r, RPTR(r), MRSIZE(m));
}

void
mpr_trace (const char *name, mpr_srcptr r, mpm_srcptr m)
{
  mp_srcptr rp;
  mp_size_t rsize;
  mpz_t  rz, getz;

  if (r == NULL)
    {
      mpz_trace (name, NULL);
      return;
    }

  rp = (mp_ptr) RPTR(r);
  if (m->_mpr_mul == mpr_2exp__mul)
    rsize = MNORM_LSHIFT(m);
  else
    rsize = MRSIZE(m);
  MPZ_FAKE_N (rz, rp, rsize);

  mp_trace_start (name);
  mpz_out_str (stdout, mp_trace_base, rz);

  /* converted out into plain number, if different from residue form */
  mpz_init (getz);
  mpr_get_z (getz, r, m);
  if (mpz_cmp (getz, rz) != 0)
    {
      printf (" (");
      mpz_out_str (stdout, mp_trace_base, getz);
      printf (")");
    }
  mpz_clear (getz);

  printf ("\n");
}

char *
mpm_name (mpm_srcptr m)
{
  if (m->_mpr_mul == mpr_2exp__mul)
    return "2exp";

  if (m->_mpr_mul == mpr_2expsub1__mul__shift
      || m->_mpr_mul == mpr_2expsub1__mul__noshift)
    return "2expsub1";

  if (m->_mpr_mul == mpr_div__mul)
    return "div";

  if (m->_mpr_mul == mpr_redc__mul_big)
    return "redc (big)";

  if (m->_mpr_mul == mpr_redc__mul_small)
    return "redc (small)";

  if (m->_mpr_mul == mpr_one__mul)
    return "one-limb";

  return "unknown";
}

void
mpm_dump (mpm_srcptr m)
{
  printf ("m %p: %s %ld %p\n", m, mpm_name(m), MRSIZE(m), MINV(m));
  mpz_trace ("  modulus  ", m->_mp_m);
  printf    ("  mbits    = %lu\n", MBITS(m));
  printf    ("  rsize    = %ld\n", MRSIZE(m));
  printf    ("  ralloc   = %ld\n", MRALLOC(m));
  mpn_trace ("  inv      ", MINV(m), MRSIZE(m));
  mpn_trace ("  smallinv ", m->_mp_smallinv, numberof (m->_mp_smallinv));
  mpn_trace ("  rmod     ", RPTR(MRMOD(m)), MRSIZE(m));
  mpn_trace ("  r2mod    ", RPTR(MR2MOD(m)), MRSIZE(m));
  mpn_trace ("  r3mod    ", RPTR(MR3MOD(m)), MRSIZE(m));
  mpn_trace ("  negone   ", RPTR(MNEGONE(m)), MRSIZE(m));
  mpn_trace ("  norm_rshift_mask", &MNORM_RSHIFT_MASK(m), (mp_size_t) 1);
  printf    ("  norm_rshift =%d\n", MNORM_RSHIFT(m));
  printf    ("  norm_lshift =%d\n", MNORM_LSHIFT(m));
  printf    ("  mbits       =%lu\n", m->_mp_mbits);
}
